import os
from uuid import uuid4
import logging
from functools import wraps, partial

from kombu import Queue, Connection, Exchange
from kombu.pools import producers
from kombu.common import send_reply
from amqp import exceptions

from .settings import CONN_DICT

class Qurator(object):

    """Manage Queue and callbacks for a set of Consumers"""

    queues = {}
    callbacks = {}
    dispatch = {}

    def __init__(self,
                 queue=None,
                 prefix='qurator',
                 exchange=Exchange(),
                 task_exchange=Exchange(),
                 conn_dict=CONN_DICT):
        """Constructor

        :prefix: Prefix for consumer queues. Defaults to 'qurator'.
        :queue: Default name for queue
        :exchange: Exchange to use.
        """
        logger = logging.getLogger(__name__)
        self.conn_dict = CONN_DICT
        self._exchange = exchange
        self._task_exchange = task_exchange
        self._prefix = prefix
        self._queue = queue
        logger.debug("Initialised Qurator class")

    def _error(self, error, message):
        """Return an error if caller sent an unknown command.

        :error: Error data
        :message: Message object

        """
        message.ack()
        self.respond_to_client(message, error)

    def _wrap_function(self, function, callback, queue_name, task=False):
        """Set up queue used in decorated function.

        :func: wrapped function
        :exchange: exchange to use
        :queue_name: name of queue
        :returns: wrapped function

        """

        logger = logging.getLogger(__name__)
        name = function.__name__.lower()
        if name not in self.queues:
            self.queues[name] = []
        if name not in self.callbacks:
            self.callbacks[name] = []
        self.callbacks[name].append(callback)
        # If not set by instance, make same as function name.
        if queue_name is None:
            queue_name = '.'.join([self._prefix, name])

        logger.debug("Setting up %s" % queue_name)
        routing_key = queue_name
        # Create the queue.
        exchange = self._exchange

        if task:
            exchange = self._task_exchange
        queue = Queue(queue_name,
                      exchange,
                      durable=exchange.durable,
                      routing_key=routing_key)

        self.queues[name].append(queue)
        # The function returned by the decorator don't really do
        # anything.  The process_msg callback added to the consumer
        # is what actually responds to messages  from the client
        # on this particular queue.

        def decorate(func):
            @wraps(func)
            def wrapper(*args, **kwargs):
                pass
            return wrapper
        return decorate

    def task(self, func=None, *, queue_name=None):
        """Wrap around a function that should be a task.
        The client should not expect anything to be returned.

        """
        logger = logging.getLogger(__name__)
        if queue_name:
            logger = logging.getLogger(queue_name)
        if not self._task_exchange.durable:
            raise Exception('Task exchange should be durable.')
        if func is None:
            return partial(self.task, queue_name=queue_name)

        def process_message(body, message):
            logger.debug("Processing function {!r} "
                         " in message: {!r} "
                         "with data {!r}".format(func.__name__,
                                                 message,
                                                 body))
            try:
                func(body)
            except Exception:
                logger.error("Problem processing task", exc_info=True)
            else:
                logger.debug("Ack'ing message.")
                message.ack()

        return self._wrap_function(
            func, process_message, queue_name, task=True)

    def rpc(self, func=None, *, queue_name=None):
        """Wrap around function. This method is modelled after standard RPC
        behaviour where the message sends a reply_to queue and a
        correlation_id back to the client.

        :func: wrap with new standard rpc behaviour
        :queue_name: defaults to "qurator.<func.__name__>"

        """
        if queue_name:
            logger = logging.getLogger(queue_name)
        else:
            logger = logging.getLogger(__name__)
        if func is None:
            return partial(self.rpc, queue_name=queue_name)

        def process_message(body, message):
            logger.debug("Processing function {!r} "
                         "with data {!r}".format(func.__name__, body))
            try:
                properties = message.properties
                correlation_id = properties.setdefault( 'correlation_id', uuid4().hex)
                logger.info('STARTSERVICE:%s;CORRELATION_ID:%s' % (queue_name,
                                                                   correlation_id))
            except Exception:
                logger.error("No correlation id for request!"
                             " {!r}".format(body),
                             exc_info=True)
            response = func(body)
            logger.info("Wrapped method returned:  {!r}".format(response))
            self.respond_to_client(message, response, queue_name)
            message.ack()

        return self._wrap_function(func, process_message, queue_name)

    def respond_to_client(self, message, response=None, queue_name=None):
        """Send RPC response back to client.

        :response: datastructure that needs to go back to client.
        """
        if response is None:
            response = {}
        if queue_name is None:
            queue_name = __name__
        logger = logging.getLogger(queue_name)
        logger.info("Replying to queue {!r} with properties: {!r}".format(
            message.properties['reply_to'],
            message.properties['correlation_id']
        ))
        with Connection(**CONN_DICT) as conn:
            with producers[conn].acquire(block=True) as producer:
                try:
                    send_reply(
                        self._exchange,
                        message,
                        response,
                        producer,
                        retry=True,
                        retry_policy={
                            'max_retries': 3,
                            'interval_start': 0,
                            'interval_step': 0.2,
                            'interval_max': 0.2,
                        }
                    )
                except exceptions.AMQPError:
                    logger.error("Problem communicating with rabbit",
                                 exc_info=True)
                except KeyError:
                    logger.error('Missing key in request', exc_info=True)
                except Exception:
                    logger.error('Unable to reply to request', exc_info=True)
                else:
                    correlation_id = message.properties['correlation_id']
                    logger.info('STOPSERVICE:%s;CORRELATION_ID:%s' % (
                        __name__, correlation_id))
                    logger.info('Replied with response {!r}'.format(response))

    def run(self):
        from .worker import Worker
        logger = logging.getLogger(__name__)
        logger.info("running worker with connection: {!r}".format(self.conn_dict))

        with Connection(**self.conn_dict) as conn:
            try:
                conn.connect()
                worker = Worker(conn, self)
                worker.run()
            except KeyboardInterrupt:
                print('bye bye')



from unittest.mock import MagicMock, ANY
from kombu import Consumer, Queue
from kombu.utils import nested

from ..queue import Qurator
from ..rpc.client import RpcClient

from .test_rabbit import TestRabbitpy


class TestAbstractMQ(TestRabbitpy):

    """Test RabbitMQ interaction"""

    def setUp(self):
        """Setup unit tests """
        super().setUp()
        self.queues = []

    def pre_declare_queues(self, queues):
        """Pre-declare any queues that will be used in tests.

        :queues: list of names of queues

        """

        declared_queues = []
        for queue_name in queues:
            q = Queue(queue_name,
                      self._exchange,
                      channel=self._connection,
                      durable=self._exchange.durable,
                      routing_key=queue_name)
            q.declare()
            declared_queues.append(q)

        self.queues = declared_queues

    def test_method_wrapping(self):
        """Test creating custom rpc endpoint."""

        self.pre_declare_queues(['qurator.moffa',
                                 'qurator.boffa',
                                 'boffa.moffa'])
        consumer = Qurator(exchange=self._exchange)

        @consumer.rpc
        def moffa(msg):
            pass

        self.assertIn('moffa', consumer.queues)
        moffa_queues = consumer.queues['moffa']
        self.assertEqual(len(moffa_queues),
                         1,
                         'One consumer')
        self.assertEqual(moffa_queues[0].name,
                         'qurator.moffa',
                         'Queue has expected name')

        @consumer.rpc(queue_name='boffa.moffa')
        def boffa(msg):
            pass

        self.assertIn('boffa', consumer.queues)
        boffa_queues = consumer.queues['boffa']
        boffa_queue = boffa_queues[0]
        self.assertEqual(boffa_queue.name,
                         'boffa.moffa',
                         'Can specify queue name')

    def test_standard_rpc(self):
        """Check behaviour of wrapped function."""

        self.pre_declare_queues(['qurator.blah', 'blah.client'])
        consumer = Qurator(exchange=self._exchange)
        checkit = MagicMock(return_value={"msg": "Got reply"})
        # Now mock it!
        consumer.respond_to_client = MagicMock()

        @consumer.rpc
        def blah(*args, **kwargs):
            return checkit(*args, **kwargs)

        payload = {"msg": "Hello"}

        # Send message to server
        client = RpcClient(exchange=self._exchange, prefix='qurator')
        reply = client.rpc('blah', payload)

        # Synthetically drain events from queues
        blah_queues = consumer.queues['blah']
        blah_callbacks = consumer.callbacks['blah']
        conn = self._connection
        with Consumer(conn, blah_queues, callbacks=blah_callbacks):
            conn.drain_events(timeout=1)

        checkit.assert_called_with(payload)
        # response = client.retrieve_messages()
        consumer.respond_to_client.assert_called_with(
            ANY,
            {"msg": "Got reply"},
            None
        )

    def test_rpc_client(self):
        """Check behaviour of client """
        self.pre_declare_queues(['qurator.booya'])
        consumer = Qurator(exchange=self._exchange)

        @consumer.rpc
        def booya(*args, **kwargs):
            return {"msg": "Wooot"}

        payload = {"msg": "Boooya"}
        client = RpcClient(exchange=self._exchange, prefix='qurator')
        client.rpc('booya', payload)
        booya_queue = consumer.queues['booya']
        booya_callbacks = consumer.callbacks['booya']

        conn = self._connection
        with Consumer(conn, booya_queue, callbacks=booya_callbacks):
            conn.drain_events(timeout=1)
        reply = client.retrieve_messages()
        self.assertIn('msg', reply)
        self.assertEqual(reply['msg'], 'Wooot')

    def test_default_exchange(self):
        """Test behaviour using the default exchange."""

        from kombu import Exchange
        self._exchange = Exchange(
            channel=self._connection,
            durable=True,
            type='direct')
        self.pre_declare_queues(['qurator.testing_default_exchange'])
        q = Qurator(exchange=self._exchange)

        @q.rpc
        def testing_default_exchange(data):
            return_data = {}
            return_data.update(data)
            return {"x": 1, "data": return_data}

        client = RpcClient(exchange=self._exchange)

        request = {'request': 'my request'}
        client.rpc('testing_default_exchange',
                   request,
                   server_routing_key='qurator.testing_default_exchange')

        queues = q.queues['testing_default_exchange']
        test_callbacks = q.callbacks['testing_default_exchange']
        conn = self._connection
        with Consumer(conn, queues, callbacks=test_callbacks):
            conn.drain_events(timeout=1)
        reply = client.retrieve_messages()
        self.assertIn('x', reply)
        self.assertEqual(reply['data'], request)

    def test_task_nondurable_exchange(self):
        """Task setup """
        from kombu import Exchange
        e = Exchange('', durable=False, type='direct')
        q = Qurator(task_exchange=e, queue='test.nondurable_exchange.queue')

        with self.assertRaises(Exception):
            @q.task
            def amation(data):
                return None

    def test_task_fail(self):
        """What happens when a task fails.
        Expected behaviour is that it does not ack and is left in the queue.
        """
        from kombu import Exchange
        e = Exchange('', type='direct')
        # declare queue
        consumer_queue = Queue('test.task.fail',
                               e,
                               channel=self._connection,
                               routing_key='test.task.fail')
        client_queue = Queue('',
                             e,
                             durable=False,
                             channel=self._connection)
        consumer_queue.declare()
        client_queue.declare()
        self.queues.append(consumer_queue)
        self.queues.append(client_queue)

        q = Qurator(task_exchange=e)

        @q.task(queue_name='test.task.fail')
        def fail(data):
            raise Exception('YOU FAIL!')

        client = RpcClient(exchange=e)
        client.task('fail', {'x': 1}, server_routing_key='test.task.fail')

        curr_queues = q.queues['fail']
        curr_callbacks = q.callbacks['fail']

        def still_around(body, message):
            self.assertFalse(message.acknowledged)
            message.ack()

        curr_callbacks.append(still_around)

        with Consumer(self._connection, curr_queues, callbacks=curr_callbacks):
            self._connection.drain_events(timeout=1)

    def test_task_succeed(self):
        """What happens when a task succeeds.

        Expect it to be acked from the queue.
        """
        from kombu import Exchange
        e = Exchange('', type='direct')
        # declare queue
        consumer_queue = Queue('test.task.succeed',
                               e,
                               channel=self._connection,
                               routing_key='test.task.succeed')
        client_queue = Queue('',
                             e,
                             durable=False,
                             channel=self._connection)
        consumer_queue.declare()
        client_queue.declare()
        self.queues.append(client_queue)

        q = Qurator(task_exchange=e, queue='test.task.succeed')

        @q.task
        def succeed(data):
            return None

        client = RpcClient(exchange=e)
        client.task('succeed',
                    {'x': 1},
                    server_routing_key='test.task.succeed')

        curr_queues = q.queues['succeed']

        @q.rpc
        def still_around(body, message):
            print("Message: {!r}".format(message))

            self._connection.drain_events(timeout=1)



import os
import sys

DEPLOY_ENVIRONMENT = os.environ.get('DEPLOY_ENVIRONMENT', 'development')
if DEPLOY_ENVIRONMENT is None:
    raise Exception("Please define the local DEPLOY_ENVIRONMENT.")
RABBITMQ_TRANSPORT_SERVICE_HOST = os.environ.get(
    'RABBITMQ_TRANSPORT_SERVICE_HOST',
    'localhost')
RABBITMQ_TRANSPORT_SERVICE_PORT = os.environ.get(
    'RABBITMQ_TRANSPORT_SERVICE_PORT',
    5672)
RABBITMQ_USER = os.environ.get('RABBITMQ_USER', 'guest')
RABBITMQ_PASSWORD = os.environ.get('RABBITMQ_PASSWORD', 'guest')
RABBITMQ_VHOST = os.environ.get('RABBITMQ_VHOST', '/')
RABBITMQ_SSL = os.environ.get('RABBITMQ_SSL', False)

for mq_key in [RABBITMQ_USER, RABBITMQ_PASSWORD, RABBITMQ_VHOST]:
    if mq_key is None:
        raise Exception('Must define RABBITMQ_USER, RABBITMQ_PASSWORD and '
                        'RABBITMQ_VHOST.')

CONN_DICT = {
    'hostname': RABBITMQ_TRANSPORT_SERVICE_HOST,
    'port': RABBITMQ_TRANSPORT_SERVICE_PORT,
    'userid': RABBITMQ_USER,
    'password': RABBITMQ_PASSWORD,
    'ssl': RABBITMQ_SSL,
    'virtual_host': RABBITMQ_VHOST,
    'heartbeat': 2
}



from unittest import TestCase

from kombu import Connection, Exchange, Queue, Consumer

from ..settings import CONN_DICT


class TestRabbitpy(TestCase):

    """Superclass for qurator testing stuff"""

    exchange_name = 'testqurator'
    conn_dict = CONN_DICT

    def setUp(self):
        """Constructor """
        self._connection = self.connection_factory()
        try:
            self._exchange = Exchange()
        except OSError as derp:
            if isinstance(derp.args[0], ConnectionRefusedError):
                self.skipTest("Not connected to RabbitMQ. Skipping.")
            raise

    def tearDown(self):
        """tear down unit tests """

        for queue in self.queues:
            if not queue.durable:
                queue.purge()
                queue.delete()
        self._connection.release()

    def connection_factory(self):
        """Return connection object
        :returns: Connection object

        """
        print("Connection: {!r}".format(self.conn_dict))
        c = Connection(**self.conn_dict)
        return c

    def pre_declare_queues(self, queues):
        """Pre-declare any queues that will be used in tests.

        :queues: list of names of queues

        """

        declared_queues = []
        for queue_name in queues:
            q = Queue(queue_name,
                      self._exchange,
                      channel=self._connection,
                      durable=self._exchange.durable,
                      routing_key=queue_name)
            q.declare()
            declared_queues.append(q)
        self.queues = declared_queues

    def pull_messages(self,
                      qurator,
                      queues=None,
                      callbacks=None,
                      command=None):
        """Helper to pull messages from a particular queue.

        :qurator: Qurator object
        :command: queue set to pull from
        :callbacks: callbacks from
        """
        if (queues is None or callbacks is None) and command is None:
            raise Exception("Unable to determine "
                            "which queue or callback to use. "
                            "Please provide either both queues and callbacks "
                            "or a command to check")
        if queues is None:
            queues = qurator.queues[command]
        if callbacks is None:
            callbacks = qurator.callbacks[command]
        with Consumer(self._connection, queues, callbacks=callbacks):
            self._connection.drain_events(timeout=1)



from unittest import TestCase
from unittest.mock import MagicMock, ANY

from qurator.rpc.client import RpcClient


class TestClient(TestCase):

    """Test the rpc client."""

    def test_payload_standard_rpc(self):
        """Test command for standard RPC."""
        c = RpcClient()
        c._send_command = MagicMock()
        c.retrieve_messages = MagicMock()
        call_payload = {"test": 2}
        c.rpc('do_that', call_payload)
        c._send_command.assert_called_with(call_payload,
                                           'do_that',  # routing key
                                           ANY)  # properties

    def test_rpc_properties(self):
        """Test setup rpc properties.  """
        c = RpcClient(client_queue='whatever.client')
        c._send_command = MagicMock()
        c.retrieve_messages = MagicMock()
        c.rpc('whatever', {"data": "x"})
        c._send_command.assert_called_with(
            ANY,
            'whatever',
            {
                "reply_to": "whatever.client",
                "correlation_id": ANY
            }
        )
        # This part shouldn't matter if legacy or not.
        c = RpcClient(client_queue='test.my.queue')
        c._send_command = MagicMock()
        c.retrieve_messages = MagicMock()
        c.rpc('whatever', {"data": "x"})
        c._send_command.assert_called_with(
            ANY,
            'whatever',
            {
                "reply_to": "test.my.queue",
                "correlation_id": ANY
            }
        )

    def test_task_properties(self):
        """Setup task properties."""
        c = RpcClient()
        c._send_command = MagicMock()
        c.task('whatever', {"data": "x"})
        c._send_command.assert_called_with(
            ANY,
            'whatever',
            declare_queue=False
        )
        # This part shouldn't matter if legacy or not.
        c = RpcClient(client_queue='test.my.queue')
        c._send_command = MagicMock()
        c.task('whatever', {"data": "x"})
        c._send_command.assert_called_with(
            ANY,
            'whatever',
            declare_queue=False
        )



import logging
from kombu import Queue, Connection
from kombu.pools import producers
from kombu.common import uuid, collect_replies
from amqp import exceptions


from ..settings import CONN_DICT
from kombu import Exchange as default_exchange

logger = logging.getLogger(__name__)


class RpcClient(object):

    """Process a RPC queue and fetch the response."""

    reply_received = False
    messages = {}
    corr_id_server_queue = {}

    def __init__(self,
                 exchange=default_exchange,
                 prefix=None,
                 client_queue=None):
        """Constructor for client object.

        :exchange: Exchange to use
        :prefix: for automatic server routing key generation
        :client_queue: name for the client queue. Default is <command>.client
        """
        self._prefix = prefix
        self._exchange = exchange
        self.reply = None
        self._client_queue = client_queue
        self._queue = None
        self._conn = Connection(**CONN_DICT)

    def retrieve_messages(self):
        """Process the message queue and ack the one that matches the
        correlation_id sent to the server.

        :correlation_id: Expected correlation id (uuid).
        :queue: The queue to read.
        :returns: JSON object retrieved from the queue.

        """

        logger.debug("Client queue: {!r}".format(self._client_queue))
        client_queue = self._queue
        logger.debug("connection is {!r}"
                     "is connected: {!r}".format(self._conn,
                                                 self._conn.connected))
        for i in collect_replies(self._conn,
                                 self._conn.channel(),
                                 client_queue,
                                 timeout=1,
                                 limit=1,
                                 callbacks=[self.ack_message]):
            logger.debug("Received {!r}".format(i))
            if self.reply is not None:
                response = self.reply
                self.reply = None
                try:
                    client_queue.purge()
                    client_queue.delete()
                    self._conn.release()
                except Exception:
                    logger.warn("Unable to purge and delete queues",
                                exc_info=True)
                return response

    def ack_message(self, body, message):
        logger.info("Processing message: {!r} with body {!r}".format(message,
                                                                     body))
        if 'correlation_id' in message.properties:
            corr_id = message.properties['correlation_id']
            try:
                self.messages.pop(corr_id)
                server_queue = self.corr_id_server_queue.pop(corr_id)
                logger.info(
                    "STOPREQUEST:%s;CORRELATION_ID:%s" %
                    (server_queue, corr_id))
                self.reply = body
                message.ack()
            except KeyError as e:
                logger.exception(e)
                logger.error("Malformed message: {!r}".format(body))

    def _setup_payload(self, command_name, data):
        """Setup the datastructure for either hase-like or standard.

        :command_name: the name of the command
        :data: data to be sent with the command
        :returns: payload for the request

        """
        return data

    def _prepare_client_queue(self, command_name):
        """Setup a client queue based on the command.

        """
        if self._client_queue is None:
            corr_uuid = uuid()
            self._client_queue = '.'.join([command_name, 'client', corr_uuid])

    def rpc(self,
            command_name,
            data=None,
            server_routing_key=None):
        """Send a RPC request

        :command_name: the command to execute (used as routing key)
        :data: dict with data to be sent
        :client_queue: Queue for this particular request
        :server_routing_key: Server routing key. Will
        default to <command>.server
        """

        self.reply_received = False
        payload = self._setup_payload(command_name, data)
        logger.info("Preparing request {!r}".format(payload))

        if server_routing_key is None:
            if self._prefix is None:
                server_routing_key = command_name
            else:
                server_routing_key = '.'.join([self._prefix, command_name])

        self._prepare_client_queue(command_name)
        logger.info("Set up client queue {!r} "
                    "to {!r}".format(self._client_queue,
                                     server_routing_key))

        message_correlation_id = uuid()
        properties = {
            'reply_to': self._client_queue,
            'correlation_id': message_correlation_id
        }
        self.corr_id_server_queue[message_correlation_id] = server_routing_key
        logger.info('STARTREQUEST:%s;CORRELATION_ID:%s' % (server_routing_key,
                                                           message_correlation_id))
        result = None
        try:
            self.messages[message_correlation_id] = True
            self._send_command(payload, server_routing_key, properties)
            result = self.retrieve_messages()
        except Exception:
            logger.error("Sending message to consumer queue failed.", 
                         exc_info=True)
            raise
        # Successful so store message correlation id for retrieval.
        return result

    def task(self,
             command_name,
             data=None,
             server_routing_key=None):
        """Send a RPC request

        :command_name: the command to execute (used as routing key)
        :data: dict with data to be sent
        :client_queue: Queue for this particular request
        :server_routing_key: Server routing key. Will default
        to <command>.server
        """

        if data is None:
            data = {}
        self.reply_received = False
        payload = self._setup_payload(command_name, data)
        logger.debug("Preparing request {!r}".format(payload))

        if server_routing_key is None:
            server_routing_key = command_name

        self._prepare_client_queue(command_name)
        logger.info("Set up client queue {!r} "
                    "to {!r}".format(self._client_queue,
                                     server_routing_key))

        self._send_command(payload, server_routing_key, declare_queue=False)

    def _send_command(self, payload, server_routing_key, properties=None,
                      declare_queue=True):
        if properties is None:
            properties = {}
        self.reply = None
        logger.debug("Using connection: {!r}".format(CONN_DICT))
        logger.info("Declaring queue %s." % self._client_queue)
        queue = Queue(self._client_queue,
                      channel=self._conn,
                      durable=self._exchange.durable,
                      exchange=self._exchange,
                      routing_key=self._client_queue)
        if declare_queue:
            queue.declare()
        self._queue = queue
        with producers[self._conn].acquire(block=True) as producer:
            producer.publish(payload,
                             serializer='json',
                             exchange=self._exchange,
                             declare=[self._exchange],
                             routing_key=server_routing_key,
                             **properties)
            logger.info("Published {!r} to exchange "
                        "{!r}".format(payload, self._exchange))
